@@ -18,6 +18,7 @@
  * 02111-1307, USA.
  *
  * Authors: William Jon McCann <mccann@jhu.edu>
+ *          Rodrigo Moya <rodrigo@novell.com>
  *
  */
 
@@ -32,8 +33,9 @@
 #include <gtk/gtk.h>
 #include <glade/glade.h>
 #include <gconf/gconf-client.h>
-#include <libgnome/gnome-help.h>
+
 #include <libgnomeui/gnome-ui-init.h>
+#include <libgnomeui/gnome-help.h> /* for gnome_help_display */
 #include <libgnomevfs/gnome-vfs-async-ops.h>
 #include <libgnomevfs/gnome-vfs-ops.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
@@ -45,21 +47,22 @@
 
 #define GLADE_XML_FILE "gnome-screensaver-preferences.glade"
 
-#define KEY_DIR          "/apps/gnome-screensaver"
-#define KEY_LOCK         KEY_DIR "/lock"
-#define KEY_MODE         KEY_DIR "/mode"
-#define KEY_BLANK_DELAY  KEY_DIR "/blank_delay"
-#define KEY_LOCK_DELAY   KEY_DIR "/lock_delay"
-#define KEY_CYCLE_DELAY  KEY_DIR "/cycle_delay"
-#define KEY_DPMS_ENABLED KEY_DIR "/dpms_enabled"
-#define KEY_DPMS_STANDBY KEY_DIR "/dpms_standby"
-#define KEY_DPMS_SUSPEND KEY_DIR "/dpms_suspend"
-#define KEY_DPMS_OFF     KEY_DIR "/dpms_off"
-#define KEY_THEMES       KEY_DIR "/themes"
+#define KEY_DIR             "/apps/gnome-screensaver"
+#define KEY_LOCK            KEY_DIR "/lock_enabled"
+#define KEY_IDLE_ACTIVATION_ENABLED         KEY_DIR "/idle_activation_enabled"
+#define KEY_MODE            KEY_DIR "/mode"
+#define KEY_ACTIVATE_DELAY  KEY_DIR "/idle_delay"
+#define KEY_LOCK_DELAY      KEY_DIR "/lock_delay"
+#define KEY_CYCLE_DELAY     KEY_DIR "/cycle_delay"
+#define KEY_DPMS_ENABLED    KEY_DIR "/dpms_enabled"
+#define KEY_DPMS_STANDBY    KEY_DIR "/dpms_standby"
+#define KEY_DPMS_SUSPEND    KEY_DIR "/dpms_suspend"
+#define KEY_DPMS_OFF        KEY_DIR "/dpms_off"
+#define KEY_THEMES          KEY_DIR "/themes"
 
 enum {
-        LABEL_COLUMN,
         NAME_COLUMN,
+        ID_COLUMN,
         N_COLUMNS
 };
 
@@ -67,7 +70,6 @@ static GConfEnumStringPair mode_enum_map [] = {
        { GS_MODE_BLANK_ONLY,       "blank-only" },
        { GS_MODE_RANDOM,           "random"     },
        { GS_MODE_SINGLE,           "single"     },
-       { GS_MODE_DONT_BLANK,       "disabled"   },
        { 0, NULL }
 };
 
@@ -87,7 +89,7 @@ static GladeXML *xml = NULL;
 static GSJob    *job = NULL;
 
 static gint32
-config_get_blank_delay (gboolean *is_writable)
+config_get_activate_delay (gboolean *is_writable)
 {
         GConfClient *client;
         gint32       delay;
@@ -96,11 +98,11 @@ config_get_blank_delay (gboolean *is_writable)
 
         if (is_writable) {
                 *is_writable = gconf_client_key_is_writable (client,
-                                                             KEY_BLANK_DELAY,
+                                                             KEY_ACTIVATE_DELAY,
                                                              NULL);
         }
 
-        delay = gconf_client_get_int (client, KEY_BLANK_DELAY, NULL);
+        delay = gconf_client_get_int (client, KEY_ACTIVATE_DELAY, NULL);
 
         if (delay < 1)
                 delay = 1;
@@ -111,13 +113,55 @@ config_get_blank_delay (gboolean *is_writable)
 }
 
 static void
-config_set_blank_delay (gint32 timeout)
+config_set_activate_delay (gint32 timeout)
 {
         GConfClient *client;
 
         client = gconf_client_get_default ();
 
-        gconf_client_set_int (client, KEY_BLANK_DELAY, timeout, NULL);
+        gconf_client_set_int (client, KEY_ACTIVATE_DELAY, timeout, NULL);
+
+        g_object_unref (client);
+}
+
+static int
+config_get_mode (gboolean *is_writable)
+{
+        GConfClient *client;
+        int          mode;
+        char        *string;
+
+        client = gconf_client_get_default ();
+
+        if (is_writable) {
+                *is_writable = gconf_client_key_is_writable (client,
+                                                             KEY_MODE,
+                                                             NULL);
+        }
+
+        string = gconf_client_get_string (client, KEY_MODE, NULL);
+        if (string) {
+                gconf_string_to_enum (mode_enum_map, string, &mode);
+                g_free (string);
+        } else {
+                mode = GS_MODE_BLANK_ONLY;
+        }
+
+        g_object_unref (client);
+
+        return mode;
+}
+
+static void
+config_set_mode (int mode)
+{
+        GConfClient *client;
+        const char  *mode_string;
+
+        client = gconf_client_get_default ();
+
+        mode_string = gconf_enum_to_string (mode_enum_map, mode);
+        gconf_client_set_string (client, KEY_MODE, mode_string, NULL);
 
         g_object_unref (client);
 }
@@ -126,7 +170,6 @@ static char *
 config_get_theme (gboolean *is_writable)
 {
         GConfClient *client;
-        char        *string;
         char        *name = NULL;
         int          mode;
 
@@ -145,17 +188,12 @@ config_get_theme (gboolean *is_writable)
                 *is_writable = can_write_theme && can_write_mode;
         }
 
-        string = gconf_client_get_string (client, KEY_MODE, NULL);
-        if (string) {
-                gconf_string_to_enum (mode_enum_map, string, &mode);
-                g_free (string);
-        } else
-                mode = GS_MODE_BLANK_ONLY;
+        mode = config_get_mode (NULL);
 
         if (mode == GS_MODE_BLANK_ONLY) {
                 name = g_strdup ("__blank-only");
-        } else if (mode == GS_MODE_DONT_BLANK) {
-                name = g_strdup ("__disabled");
+        } else if (mode == GS_MODE_RANDOM) {
+                name = g_strdup ("__random");
         } else {
                 GSList *list;
                 list = gconf_client_get_list (client,
@@ -177,27 +215,47 @@ config_get_theme (gboolean *is_writable)
         return name;
 }
 
+static GSList *
+get_all_theme_ids (GSJob *job)
+{
+        GSList *ids = NULL;
+        GSList *entries;
+        GSList *l;
+
+        entries = gs_job_get_theme_info_list (job);
+        for (l = entries; l; l = l->next) {
+                GSJobThemeInfo *info = l->data;
+
+                ids = g_slist_prepend (ids, g_strdup (gs_job_theme_info_get_id (info)));
+                gs_job_theme_info_unref (info);
+        }
+        g_slist_free (entries);
+
+        return ids;
+}
+
 static void
-config_set_theme (const char *name)
+config_set_theme (const char *theme_id)
 {
         GConfClient *client;
         GSList      *list = NULL;
-        const char  *mode_string;
         int          mode;
 
         client = gconf_client_get_default ();
 
-        if (name && strcmp (name, "__disabled") == 0) {
-                mode = GS_MODE_DONT_BLANK;
-        } else if (name && strcmp (name, "__blank-only") == 0) {
+        if (theme_id && strcmp (theme_id, "__blank-only") == 0) {
                 mode = GS_MODE_BLANK_ONLY;
+        } else if (theme_id && strcmp (theme_id, "__random") == 0) {
+                mode = GS_MODE_RANDOM;
+
+                /* set the themes key to contain all available screensavers */
+                list = get_all_theme_ids (job);
         } else {
                 mode = GS_MODE_SINGLE;
-                list = g_slist_append (list, g_strdup (name));
+                list = g_slist_append (list, g_strdup (theme_id));
         }
 
-        mode_string = gconf_enum_to_string (mode_enum_map, mode);
-        gconf_client_set_string (client, KEY_MODE, mode_string, NULL);
+        config_set_mode (mode);
 
         gconf_client_set_list (client,
                                KEY_THEMES,
@@ -205,12 +263,45 @@ config_set_theme (const char *name)
                                list,
                                NULL);
 
-        g_slist_foreach (list, (GFunc)g_free, NULL);
+        g_slist_foreach (list, (GFunc) g_free, NULL);
         g_slist_free (list);
 
         g_object_unref (client);
 }
 
+static gboolean
+config_get_enabled (gboolean *is_writable)
+{
+        int          enabled;
+        GConfClient *client;
+
+        client = gconf_client_get_default ();
+
+        if (is_writable) {
+                *is_writable = gconf_client_key_is_writable (client,
+                                                             KEY_LOCK,
+                                                             NULL);
+        }
+
+        enabled = gconf_client_get_bool (client, KEY_IDLE_ACTIVATION_ENABLED, NULL);
+
+        g_object_unref (client);
+
+        return enabled;
+}
+
+static void
+config_set_enabled (gboolean enabled)
+{
+        GConfClient *client;
+
+        client = gconf_client_get_default ();
+
+        gconf_client_set_bool (client, KEY_IDLE_ACTIVATION_ENABLED, enabled, NULL);
+
+        g_object_unref (client);
+}
+
 static gboolean
 config_get_lock (gboolean *is_writable)
 {
@@ -257,18 +348,7 @@ static void
 preview_set_theme (GtkWidget  *widget,
                    const char *theme)
 {
-        GtkWidget *delay_box;
-        GtkWidget *lock_box;
-        gboolean   delay_writable;
-        gboolean   lock_writable;
-        GError    *error = NULL;
-
-        delay_box = glade_xml_get_widget (xml, "blank_delay_hbox");
-        lock_box = glade_xml_get_widget (xml, "lock_checkbox");
-
-        /* see if the keys are writable */
-        config_get_blank_delay (&delay_writable);
-        config_get_lock (&lock_writable);
+        GError *error = NULL;
 
         if (job) {
                 gs_job_stop (job);
@@ -276,23 +356,27 @@ preview_set_theme (GtkWidget  *widget,
 
         preview_clear (widget);
 
-        if (theme && strcmp (theme, "__disabled") == 0) {
-                if (delay_writable)
-                        gtk_widget_set_sensitive (delay_box, FALSE);
-                if (lock_writable)
-                        gtk_widget_set_sensitive (lock_box, FALSE);
-                
-        } else if (theme && strcmp (theme, "__blank-only") == 0) {
-                if (delay_writable)
-                        gtk_widget_set_sensitive (delay_box, TRUE);
-                if (lock_writable)
-                        gtk_widget_set_sensitive (lock_box, TRUE);
-        } else {
-                if (delay_writable)
-                        gtk_widget_set_sensitive (delay_box, TRUE);
-                if (lock_writable)
-                        gtk_widget_set_sensitive (lock_box, TRUE);
+        if ((theme && strcmp (theme, "__blank-only") == 0)) {
 
+        } else if (theme && strcmp (theme, "__random") == 0) {
+                GSList *themes;
+
+                themes = get_all_theme_ids (job);
+                if (themes) {
+                        if (! gs_job_set_theme (job, (const char *) themes->data, &error)) {
+                                if (error) {
+                                        g_warning ("Could not set theme: %s", error->message);
+                                        g_error_free (error);
+                                }
+                                return;
+                        }
+                        g_slist_foreach (themes, (GFunc) g_free, NULL);
+                        g_slist_free (themes);
+
+                        gs_job_start (job);
+                }
+                
+        } else {
                 if (! gs_job_set_theme (job, theme, &error)) {
                         if (error) {
                                 g_warning ("Could not set theme: %s", error->message);
@@ -309,71 +393,77 @@ response_cb (GtkWidget *widget,
              int        response_id)
 {
 
-        if (response_id == GTK_RESPONSE_HELP)
+        if (response_id == GTK_RESPONSE_HELP) {
+                GError *error;
+                error = NULL;
+                gnome_help_display_desktop (NULL,
+                                            "user-guide",
+                                            "user-guide.xml",
+                                            "prefs-screensaver",
+                                            &error);
+                if (error != NULL) {
+                        g_warning ("%s", error->message);
+                        g_error_free (error);
+                }
+
                 return;
+        }
 
         gtk_widget_destroy (widget);
         gtk_main_quit ();
 }
 
 static GSList *
-get_theme_list (void)
+get_theme_info_list (void)
 {
-        return gs_job_get_theme_list (job);
+        return gs_job_get_theme_info_list (job);
 }
 
 static void
 populate_model (GtkTreeStore *store)
 {
         GtkTreeIter iter;
-        gboolean    show_disabled = TRUE;
         GSList     *themes        = NULL;
         GSList     *l;
 
-        if (show_disabled) {
-                gtk_tree_store_append (store, &iter, NULL);
-                gtk_tree_store_set (store, &iter,
-                                    LABEL_COLUMN, "Disabled",
-                                    NAME_COLUMN, "__disabled",
-                                    -1);
-        }
-
         gtk_tree_store_append (store, &iter, NULL);
         gtk_tree_store_set (store, &iter,
-                            LABEL_COLUMN, "Blank screen",
-                            NAME_COLUMN, "__blank-only",
+                            NAME_COLUMN, _("Blank screen"),
+                            ID_COLUMN, "__blank-only",
                             -1);
 
         gtk_tree_store_append (store, &iter, NULL);
         gtk_tree_store_set (store, &iter,
-                            LABEL_COLUMN, NULL,
-                            NAME_COLUMN, "__separator",
+                            NAME_COLUMN, _("Random"),
+                            ID_COLUMN, "__random",
                             -1);
 
-        themes = get_theme_list ();
+        gtk_tree_store_append (store, &iter, NULL);
+        gtk_tree_store_set (store, &iter,
+                            NAME_COLUMN, NULL,
+                            ID_COLUMN, "__separator",
+                            -1);
+
+        themes = get_theme_info_list ();
 
         if (! themes)
                 return;
         
         for (l = themes; l; l = l->next) {
-                GSJobThemeInfo *info;
-
-                info = gs_job_lookup_theme_info (job,
-                                                 (const char *) l->data);
+                GSJobThemeInfo *info = l->data;
 
                 if (! info)
                         continue;
 
                 gtk_tree_store_append (store, &iter, NULL);
                 gtk_tree_store_set (store, &iter,
-                                    NAME_COLUMN, info->name,
-                                    LABEL_COLUMN, info->title,
+                                    NAME_COLUMN, gs_job_theme_info_get_name (info),
+                                    ID_COLUMN, gs_job_theme_info_get_id (info),
                                     -1);
 
-                gs_job_theme_info_free (info);
+                gs_job_theme_info_unref (info);
         }
 
-        g_slist_foreach (themes, (GFunc)g_free, NULL);
         g_slist_free (themes);
 }
 
@@ -388,7 +478,7 @@ tree_selection_changed_cb (GtkTreeSelection *selection,
         if (! gtk_tree_selection_get_selected (selection, &model, &iter))
                 return;
 
-        gtk_tree_model_get (model, &iter, NAME_COLUMN, &theme, -1);
+        gtk_tree_model_get (model, &iter, ID_COLUMN, &theme, -1);
 
         if (! theme)
                 return;
@@ -400,13 +490,13 @@ tree_selection_changed_cb (GtkTreeSelection *selection,
 }
 
 static void
-blank_delay_value_changed_cb (GtkRange *range,
-                              gpointer  user_data)
+activate_delay_value_changed_cb (GtkRange *range,
+                                 gpointer  user_data)
 {
         gdouble value;
 
         value = gtk_range_get_value (range);
-        config_set_blank_delay ((gint32)value);
+        config_set_activate_delay ((gint32)value);
 }
 
 static int
@@ -417,44 +507,44 @@ compare_theme  (GtkTreeModel *model,
 {
         char *name_a;
         char *name_b;
-        char *label_a;
-        char *label_b;
+        char *id_a;
+        char *id_b;
         int   result;
 
         gtk_tree_model_get (model, a, NAME_COLUMN, &name_a, -1);
         gtk_tree_model_get (model, b, NAME_COLUMN, &name_b, -1);
-        gtk_tree_model_get (model, a, LABEL_COLUMN, &label_a, -1);
-        gtk_tree_model_get (model, b, LABEL_COLUMN, &label_b, -1);
+        gtk_tree_model_get (model, a, ID_COLUMN, &id_a, -1);
+        gtk_tree_model_get (model, b, ID_COLUMN, &id_b, -1);
+
+        if (! id_a)
+                return 1;
+        else if (! id_b)
+                return -1;
+
+        if (strcmp (id_a, "__blank-only") == 0)
+                return -1;
+        else if (strcmp (id_b, "__blank-only") == 0)
+                return 1;
+        else if (strcmp (id_a, "__random") == 0)
+                return -1;
+        else if (strcmp (id_b, "__random") == 0)
+                return 1;
+        else if (strcmp (id_a, "__separator") == 0)
+                return -1;
+        else if (strcmp (id_b, "__separator") == 0)
+                return 1;
 
         if (! name_a)
                 return 1;
         else if (! name_b)
                 return -1;
 
-        if (strcmp (name_a, "__disabled") == 0)
-                return -1;
-        else if (strcmp (name_b, "__disabled") == 0)
-                return 1;
-        else if (strcmp (name_a, "__blank-only") == 0)
-                return -1;
-        else if (strcmp (name_b, "__blank-only") == 0)
-                return 1;
-        else if (strcmp (name_a, "__separator") == 0)
-                return -1;
-        else if (strcmp (name_b, "__separator") == 0)
-                return 1;
+        result = strcmp (name_a, name_b);
 
-        if (! label_a)
-                return 1;
-        else if (! label_b)
-                return -1;
-
-        result = strcmp (label_a, label_b);
-
-        g_free (label_a);
-        g_free (label_b);
         g_free (name_a);
         g_free (name_b);
+        g_free (id_a);
+        g_free (id_b);
 
         return result;
 }
@@ -498,7 +588,7 @@ setup_treeview (GtkWidget *tree,
 
         renderer = gtk_cell_renderer_text_new ();
         column = gtk_tree_view_column_new_with_attributes ("Name", renderer,
-                                                           "text", LABEL_COLUMN,
+                                                           "text", NAME_COLUMN,
                                                            NULL);
         gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);
 
@@ -513,7 +603,7 @@ setup_treeview (GtkWidget *tree,
 
         gtk_tree_view_set_row_separator_func (GTK_TREE_VIEW (tree),
                                               separator_func,
-                                              GINT_TO_POINTER (NAME_COLUMN),
+                                              GINT_TO_POINTER (ID_COLUMN),
                                               NULL);
 
         select = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
@@ -541,12 +631,12 @@ setup_treeview_selection (GtkWidget *tree)
         model = gtk_tree_view_get_model (GTK_TREE_VIEW (tree));
 
         if (theme && gtk_tree_model_get_iter_first (model, &iter)) {
-                char *name;
+                char *id;
 
                 do {
                         gtk_tree_model_get (model, &iter,
-                                            NAME_COLUMN, &name, -1);
-                        if (name && strcmp (name, theme) == 0) {
+                                            ID_COLUMN, &id, -1);
+                        if (id && strcmp (id, theme) == 0) {
                                 path = gtk_tree_model_get_path (model, &iter);
                                 break;
                         }
@@ -624,7 +714,6 @@ theme_installer_run (GtkWidget *parent,
 
         while (TRUE) {
                 char      *file_tmp;
-                GtkWidget *dialog;
                 int        len = strlen (base);
 
                 if (base && len > 4 && (!strcmp (base + len - 4, ".xml"))) {
@@ -813,20 +902,131 @@ lock_checkbox_toggled (GtkToggleButton *button, gpointer user_data)
         config_set_lock (gtk_toggle_button_get_active (button));
 }
 
+static void
+enabled_checkbox_toggled (GtkToggleButton *button, gpointer user_data)
+{
+        config_set_enabled (gtk_toggle_button_get_active (button));
+}
+
+static void
+invalid_type_warning (const char *type)
+{
+        g_warning ("Error retrieving configuration key '%s': Invalid type",
+                   type);
+}
+
+static void
+ui_set_lock (gboolean enabled)
+{
+        GtkWidget *widget;
+        gboolean   active;
+
+        widget = glade_xml_get_widget (xml, "lock_checkbox");
+
+        active = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
+        if (active != enabled) {
+                gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), enabled);
+        }
+}
+
+static void
+ui_set_enabled (gboolean enabled)
+{
+        GtkWidget *widget;
+        gboolean   active;
+        gboolean   is_writable;
+
+        widget = glade_xml_get_widget (xml, "enable_checkbox");
+        active = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
+        if (active != enabled) {
+                gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), enabled);
+        }
+
+        widget = glade_xml_get_widget (xml, "lock_checkbox");
+        config_get_lock (&is_writable);
+        if (is_writable) {
+                gtk_widget_set_sensitive (widget, enabled);
+        }
+}
+
+static void
+ui_set_delay (int delay)
+{
+        GtkWidget *widget;
+
+        widget = glade_xml_get_widget (xml, "activate_delay_hscale");
+        gtk_range_set_value (GTK_RANGE (widget), delay);
+}
+
+static void
+key_changed_cb (GConfClient *client,
+                guint        cnxn_id,
+                GConfEntry  *entry,
+                gpointer     data)
+{
+        const char *key;
+        GConfValue *value;
+
+        key = gconf_entry_get_key (entry);
+
+        if (! g_str_has_prefix (key, KEY_DIR))
+                return;
+
+        value = gconf_entry_get_value (entry);
+
+        if (strcmp (key, KEY_IDLE_ACTIVATION_ENABLED) == 0) {
+                if (value->type == GCONF_VALUE_BOOL) {
+                        gboolean enabled;
+
+                        enabled = gconf_value_get_bool (value);
+
+                        ui_set_enabled (enabled);
+                } else {
+                        invalid_type_warning (key);
+                }
+        } else if (strcmp (key, KEY_LOCK) == 0) {
+                if (value->type == GCONF_VALUE_BOOL) {
+                        gboolean enabled;
+
+                        enabled = gconf_value_get_bool (value);
+
+                        ui_set_lock (enabled);
+                } else {
+                        invalid_type_warning (key);
+                }
+        } else if (strcmp (key, KEY_ACTIVATE_DELAY) == 0) {
+
+                if (value->type == GCONF_VALUE_INT) {
+                        int delay;
+
+                        delay = gconf_value_get_int (value);
+                        ui_set_delay (delay);
+                } else {
+                        invalid_type_warning (key);
+                }
+
+        } else {
+                /*g_warning ("Config key not handled: %s", key);*/
+        }
+}
+
 static void
 init_capplet (void)
 {
         GtkWidget *dialog;
         GtkWidget *preview;
         GtkWidget *treeview;
-        GtkWidget *blank_delay_hscale;
-        GtkWidget *blank_delay_hbox;
+        GtkWidget *activate_delay_hscale;
+        GtkWidget *activate_delay_hbox;
         GtkWidget *label;
+        GtkWidget *enabled_checkbox;
         GtkWidget *lock_checkbox;
         char      *glade_file;
-        char      *path;
-        gdouble    blank_delay;
+        char      *string;
+        gdouble    activate_delay;
+        gboolean   enabled;
         gboolean   is_writable;
+        GConfClient *client;
 
         glade_file = g_build_filename (GLADEDIR, GLADE_XML_FILE, NULL);
         xml = glade_xml_new (glade_file, NULL, PACKAGE);
@@ -851,20 +1051,21 @@ init_capplet (void)
         preview            = glade_xml_get_widget (xml, "preview_area");
         dialog             = glade_xml_get_widget (xml, "prefs_dialog");
         treeview           = glade_xml_get_widget (xml, "savers_treeview");
-        blank_delay_hscale = glade_xml_get_widget (xml, "blank_delay_hscale");
-        blank_delay_hbox   = glade_xml_get_widget (xml, "blank_delay_hbox");
+        activate_delay_hscale = glade_xml_get_widget (xml, "activate_delay_hscale");
+        activate_delay_hbox   = glade_xml_get_widget (xml, "activate_delay_hbox");
+        enabled_checkbox   = glade_xml_get_widget (xml, "enable_checkbox");
         lock_checkbox      = glade_xml_get_widget (xml, "lock_checkbox");
 
-        label              = glade_xml_get_widget (xml, "blank_delay_label");
-        gtk_label_set_mnemonic_widget (GTK_LABEL (label), blank_delay_hscale);
+        label              = glade_xml_get_widget (xml, "activate_delay_label");
+        gtk_label_set_mnemonic_widget (GTK_LABEL (label), activate_delay_hscale);
         label              = glade_xml_get_widget (xml, "savers_label");
         gtk_label_set_mnemonic_widget (GTK_LABEL (label), treeview);
 
-        blank_delay = config_get_blank_delay (&is_writable);
-        gtk_range_set_value (GTK_RANGE (blank_delay_hscale), blank_delay);
+        activate_delay = config_get_activate_delay (&is_writable);
+        ui_set_delay (activate_delay);
         if (! is_writable)
-                gtk_widget_set_sensitive (blank_delay_hbox, FALSE);
-        g_signal_connect (blank_delay_hscale, "format-value",
+                gtk_widget_set_sensitive (activate_delay_hbox, FALSE);
+        g_signal_connect (activate_delay_hscale, "format-value",
                           G_CALLBACK (format_value_callback_time), NULL);
 
         gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (lock_checkbox), config_get_lock (&is_writable));
@@ -872,6 +1073,14 @@ init_capplet (void)
                 gtk_widget_set_sensitive (lock_checkbox, FALSE);
         g_signal_connect (lock_checkbox, "toggled",
                           G_CALLBACK (lock_checkbox_toggled), NULL);
+
+        enabled = config_get_enabled (&is_writable);
+        ui_set_enabled (enabled);
+        if (! is_writable)
+                gtk_widget_set_sensitive (enabled_checkbox, FALSE);
+        g_signal_connect (enabled_checkbox, "toggled",
+                          G_CALLBACK (enabled_checkbox_toggled), NULL);
+
             
         gtk_window_set_icon_name (GTK_WINDOW (dialog), "screensaver");
 
@@ -889,21 +1098,43 @@ init_capplet (void)
 
         gtk_widget_show_all (dialog);
 
+        /* Update list of themes if using random screensaver */
+        client = gconf_client_get_default ();
+        string = gconf_client_get_string (client, KEY_MODE, NULL);
+        if (string) {
+                int mode;
+                GSList *list;
+
+                gconf_string_to_enum (mode_enum_map, string, &mode);
+                g_free (string);
+
+                if (mode == GS_MODE_RANDOM) {
+                        list = get_all_theme_ids (job);
+                        gconf_client_set_list (client, KEY_THEMES, GCONF_VALUE_STRING, list, NULL);
+
+                        g_slist_foreach (list, (GFunc) g_free, NULL);
+                        g_slist_free (list);
+                }
+        }
+
+	gconf_client_add_dir (client, KEY_DIR,
+			      GCONF_CLIENT_PRELOAD_ONELEVEL,
+                              NULL);
+	gconf_client_notify_add (client,
+                                 KEY_DIR,
+				 key_changed_cb,
+                                 NULL, NULL, NULL);
+
+        g_object_unref (client);
+
         preview_clear (preview);
         gs_job_set_widget (job, preview);
 
-        /* Add user configuration path */
-        /* FIXME: disable this if locked down */
-        path = g_build_filename (g_get_user_data_dir (), "gnome-screensaver", "themes", NULL);
-
-        gs_job_prepend_theme_path (job, path);
-        g_free (path);
-
         setup_treeview (treeview, preview);
         setup_treeview_selection (treeview);
 
-        g_signal_connect (blank_delay_hscale, "value-changed",
-                          G_CALLBACK (blank_delay_value_changed_cb), NULL);
+        g_signal_connect (activate_delay_hscale, "value-changed",
+                          G_CALLBACK (activate_delay_value_changed_cb), NULL);
 
         g_signal_connect (dialog, "response",
                           G_CALLBACK (response_cb), NULL);
@@ -922,8 +1153,11 @@ main (int    argc,
         textdomain (GETTEXT_PACKAGE);
 #endif 
 
-        gnome_program_init (PACKAGE, VERSION, LIBGNOMEUI_MODULE, argc, argv,
-                            GNOME_PARAM_NONE);
+        gnome_program_init (PACKAGE, VERSION,
+                            LIBGNOMEUI_MODULE,
+                            argc, argv,
+                            GNOME_PROGRAM_STANDARD_PROPERTIES,
+                            NULL);
 
         job = gs_job_new ();
 
