@@ -43,7 +43,7 @@ static void gs_manager_finalize   (GObject        *object);
 struct GSManagerPrivate
 {
         GSList      *windows;
-        GSList      *jobs;
+        GHashTable  *jobs;
 
         /* Policy */
         glong        lock_timeout;
@@ -98,6 +98,119 @@ static guint         signals [LAST_SIGNAL] = { 0, };
 
 G_DEFINE_TYPE (GSManager, gs_manager, G_TYPE_OBJECT)
 
+static void
+manager_add_job_for_window (GSManager *manager,
+                            GSWindow  *window,
+                            GSJob     *job)
+{
+        g_hash_table_insert (manager->priv->jobs, window, job);
+}
+
+static const char *
+select_theme (GSManager *manager)
+{
+        const char *theme = NULL;
+
+        g_return_val_if_fail (manager != NULL, NULL);
+        g_return_val_if_fail (GS_IS_MANAGER (manager), NULL);
+
+        if (manager->priv->saver_mode == GS_MODE_BLANK_ONLY) {
+                return NULL;
+        }
+
+        if (manager->priv->themes) {
+                int number = 0;
+
+                if (manager->priv->saver_mode == GS_MODE_RANDOM) {
+                        g_random_set_seed (time (NULL));
+                        number = g_random_int_range (0, g_slist_length (manager->priv->themes));
+                }
+                theme = g_slist_nth_data (manager->priv->themes, number);
+        }
+
+        return theme;
+}
+
+static void
+cycle_job (GSWindow  *window,
+           GSJob     *job,
+           GSManager *manager)
+{
+        const char *theme;
+
+        theme = select_theme (manager);
+
+        gs_job_stop (job);
+        gs_job_set_theme (job, theme, NULL);
+        gs_job_start (job);
+}
+
+static void
+manager_cycle_jobs (GSManager *manager)
+{
+        g_hash_table_foreach (manager->priv->jobs, (GHFunc) cycle_job, manager);
+}
+
+static void
+throttle_job (GSWindow  *window,
+              GSJob     *job,
+              GSManager *manager)
+{
+        if (manager->priv->throttle_enabled) {
+                gs_job_stop (job);
+        } else {
+                gs_job_start (job);
+        }
+}
+
+static void
+manager_throttle_jobs (GSManager *manager)
+{
+        g_hash_table_foreach (manager->priv->jobs, (GHFunc) throttle_job, manager);
+}
+
+static void
+resume_job (GSWindow  *window,
+            GSJob     *job,
+            GSManager *manager)
+{
+        if (gs_job_is_running (job)) {
+                gs_job_suspend (job, FALSE);
+        } else {
+                gs_job_start (job);
+        }
+}
+
+static void
+manager_resume_jobs (GSManager *manager)
+{
+        g_hash_table_foreach (manager->priv->jobs, (GHFunc) resume_job, manager);
+}
+
+static void
+suspend_job (GSWindow  *window,
+             GSJob     *job,
+             GSManager *manager)
+{
+        gs_job_suspend (job, TRUE);
+}
+
+static void
+manager_suspend_jobs (GSManager *manager)
+{
+        g_hash_table_foreach (manager->priv->jobs, (GHFunc) suspend_job, manager);
+}
+
+static void
+manager_stop_jobs (GSManager *manager)
+{
+        if (manager->priv->jobs) {
+                g_hash_table_destroy (manager->priv->jobs);
+
+        }
+        manager->priv->jobs = NULL;
+}
+
 void
 gs_manager_set_mode (GSManager  *manager,
                      GSSaverMode mode)
@@ -139,13 +252,7 @@ gs_manager_set_throttle_enabled (GSManager *manager,
 
                 if (! manager->priv->dialog_up) {
 
-                        for (l = manager->priv->jobs; l; l = l->next) {
-                                if (throttle_enabled) {
-                                        gs_job_stop (GS_JOB (l->data));
-                                } else {
-                                        gs_job_start (GS_JOB (l->data));
-                                }
-                        }
+                        manager_throttle_jobs (manager);
 
                         for (l = manager->priv->windows; l; l = l->next) {
                                 gs_window_clear (l->data);
@@ -164,8 +271,9 @@ gs_manager_set_lock_active (GSManager *manager,
                 GSList *l;
 
                 manager->priv->lock_active = lock_active;
-                for (l = manager->priv->windows; l; l = l->next)
+                for (l = manager->priv->windows; l; l = l->next) {
                         gs_window_set_lock_enabled (l->data, lock_active);
+                }
         }
 }
 
@@ -190,8 +298,9 @@ gs_manager_set_logout_enabled (GSManager *manager,
                 GSList *l;
 
                 manager->priv->logout_enabled = logout_enabled;
-                for (l = manager->priv->windows; l; l = l->next)
+                for (l = manager->priv->windows; l; l = l->next) {
                         gs_window_set_logout_enabled (l->data, logout_enabled);
+                }
         }
 }
 
@@ -205,8 +314,9 @@ gs_manager_set_user_switch_enabled (GSManager *manager,
                 GSList *l;
 
                 manager->priv->user_switch_enabled = user_switch_enabled;
-                for (l = manager->priv->windows; l; l = l->next)
+                for (l = manager->priv->windows; l; l = l->next) {
                         gs_window_set_user_switch_enabled (l->data, user_switch_enabled);
+                }
         }
 }
 
@@ -276,8 +386,9 @@ gs_manager_set_logout_timeout (GSManager *manager,
                 GSList *l;
 
                 manager->priv->logout_timeout = logout_timeout;
-                for (l = manager->priv->windows; l; l = l->next)
+                for (l = manager->priv->windows; l; l = l->next) {
                         gs_window_set_logout_timeout (l->data, logout_timeout);
+                }
         }
 }
 
@@ -296,61 +407,27 @@ gs_manager_set_logout_command (GSManager  *manager,
         }
 }
 
-static const char *
-select_theme (GSManager *manager)
-{
-        const char *theme = NULL;
-
-        g_return_val_if_fail (manager != NULL, NULL);
-        g_return_val_if_fail (GS_IS_MANAGER (manager), NULL);
-
-        if (manager->priv->saver_mode == GS_MODE_BLANK_ONLY)
-                return NULL;
-
-        if (manager->priv->themes) {
-                int number = 0;
-
-                if (manager->priv->saver_mode == GS_MODE_RANDOM) {
-                        g_random_set_seed (time (NULL));
-                        number = g_random_int_range (0, g_slist_length (manager->priv->themes));
-                }
-                theme = g_slist_nth_data (manager->priv->themes, number);
-        }
-
-        return theme;
-}
-
 gboolean
 gs_manager_cycle (GSManager *manager)
 {
-        GSList     *l;
-        const char *theme;
-        gboolean    success = FALSE;
-
         g_return_val_if_fail (manager != NULL, FALSE);
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
-        if (! manager->priv->active)
+        if (! manager->priv->active) {
                 return FALSE;
-
-        if (manager->priv->dialog_up)
-                return FALSE;
-
-        if (manager->priv->throttle_enabled)
-                return FALSE;
-
-        theme = select_theme (manager);
-
-        for (l = manager->priv->jobs; l; l = l->next) {
-                gs_job_stop (GS_JOB (l->data));
-                gs_job_set_theme (GS_JOB (l->data), theme, NULL);
-
-                /* this success flag is kinda bogus */
-                if (gs_job_start (GS_JOB (l->data)))
-                        success = TRUE;
         }
 
-        return success;
+        if (manager->priv->dialog_up) {
+                return FALSE;
+        }
+
+        if (manager->priv->throttle_enabled) {
+                return FALSE;
+        }
+
+        manager_cycle_jobs (manager);
+
+        return TRUE;
 }
 
 static gboolean
@@ -359,8 +436,9 @@ cycle_timeout (GSManager *manager)
         g_return_val_if_fail (manager != NULL, FALSE);
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
-        if (! manager->priv->dialog_up)
+        if (! manager->priv->dialog_up) {
                 gs_manager_cycle (manager);
+        }
 
         return TRUE;
 }
@@ -633,12 +711,7 @@ gs_manager_finalize (GObject *object)
 
         gs_grab_release_keyboard_and_mouse ();
 
-        for (l = manager->priv->jobs; l; l = l->next) {
-                gs_job_stop (GS_JOB (l->data));
-                g_object_unref (l->data);
-        }
-        g_slist_free (manager->priv->jobs);
-        manager->priv->jobs = NULL;
+        manager_stop_jobs (manager);
 
         for (l = manager->priv->windows; l; l = l->next) {
                 gs_window_destroy (l->data);
@@ -689,23 +762,26 @@ window_dialog_up_cb (GSWindow  *window,
 
         manager->priv->dialog_up = TRUE;
 
-        /* Grab keyboard so dialog can be used */
+        /* Move keyboard and mouse grabs so dialog can be used */
         gs_grab_window (gs_window_get_gdk_window (window),
                         gs_window_get_screen (window),
                         FALSE);
 
+        /* Release the pointer grab while dialog is up so that
+           the dialog can be used.  We'll regrab it when the dialog goes down. */
+        gs_grab_release_mouse ();
+
         /* Make all other windows insensitive so we don't get events */
         for (l = manager->priv->windows; l; l = l->next) {
-                if (l->data != window)
+                if (l->data != window) {
                         gtk_widget_set_sensitive (GTK_WIDGET (l->data), FALSE);
+                }
         }
 
         if (! manager->priv->throttle_enabled) {
                 gs_debug ("Suspending jobs");
 
-                for (l = manager->priv->jobs; l; l = l->next) {
-                        gs_job_suspend (GS_JOB (l->data), TRUE);
-                }
+                manager_suspend_jobs (manager);
         }
 }
 
@@ -718,75 +794,122 @@ window_dialog_down_cb (GSWindow  *window,
         g_return_if_fail (manager != NULL);
         g_return_if_fail (GS_IS_MANAGER (manager));
 
+        /* Regrab the mouse */
+        gs_grab_window (gs_window_get_gdk_window (window),
+                        gs_window_get_screen (window),
+                        FALSE);
+
         /* Make all windows sensitive so we get events */
         for (l = manager->priv->windows; l; l = l->next) {
                 gtk_widget_set_sensitive (GTK_WIDGET (l->data), TRUE);
         }
 
         if (! manager->priv->throttle_enabled) {
-                for (l = manager->priv->jobs; l; l = l->next) {
-                        if (gs_job_is_running (GS_JOB (l->data))) {
-                                gs_job_suspend (GS_JOB (l->data), FALSE);
-                        } else {
-                                gs_job_start (GS_JOB (l->data));
-                        }
-                }
+                manager_resume_jobs (manager);
         }
 
         manager->priv->dialog_up = FALSE;
 }
 
+static void
+manager_maybe_start_job_for_window (GSManager *manager,
+                                    GSWindow  *window)
+{
+        GSJob *job;
+
+        job = g_hash_table_lookup (manager->priv->jobs, window);
+
+        if (job == NULL) {
+                gs_debug ("Job not found for window");
+                return;
+        }
+
+        if (! manager->priv->throttle_enabled) {
+                gs_job_start (job);
+        }
+}
+
+static gboolean
+manager_maybe_grab_window (GSManager *manager,
+                           GSWindow  *window)
+{
+        GdkDisplay *display;
+        GdkScreen  *screen;
+        int         monitor;
+        int         x, y;
+        gboolean    grabbed;
+
+        display = gdk_display_get_default ();
+        gdk_display_get_pointer (display, &screen, &x, &y, NULL);
+        monitor = gdk_screen_get_monitor_at_point (screen, x, y);
+
+        gdk_flush ();
+        grabbed = FALSE;
+        if (gs_window_get_screen (window) == screen
+            && gs_window_get_monitor (window) == monitor) {
+                gs_grab_window (gs_window_get_gdk_window (window),
+                                gs_window_get_screen (window),
+                                FALSE);
+                grabbed = TRUE;
+        }
+
+        return grabbed;
+}
+
+static void
+window_grab_broken_cb (GSWindow           *window,
+                       GdkEventGrabBroken *event,
+                       GSManager          *manager)
+{
+        gs_debug ("GRAB BROKEN!");
+        if (event->keyboard) {
+                gs_grab_keyboard_reset ();
+        } else {
+                gs_grab_mouse_reset ();
+        }
+}
+
 static gboolean
 window_map_event_cb (GSWindow  *window,
                      GdkEvent  *event,
                      GSManager *manager)
 {
-        GdkDisplay *display;
-        GdkScreen  *screen;
-        int         monitor;
-        int         x, y;
+        gs_debug ("Handling window map_event event");
 
-        g_return_val_if_fail (manager != NULL, FALSE);
-        g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
-        g_return_val_if_fail (window != NULL, FALSE);
-        g_return_val_if_fail (GS_IS_WINDOW (window), FALSE);
+        manager_maybe_grab_window (manager, window);
 
-        display = gdk_display_get_default ();
-        gdk_display_get_pointer (display, &screen, &x, &y, NULL);
-        monitor = gdk_screen_get_monitor_at_point (screen, x, y);
-
-        if (gs_window_get_screen (window) == screen
-            && gs_window_get_monitor (window) == monitor) {
-                gs_grab_window (gs_window_get_gdk_window (window),
-                                gs_window_get_screen (window),
-                                FALSE);
-        }
+        manager_maybe_start_job_for_window (manager, window);
 
         return FALSE;
 }
 
 static void
-window_show_cb (GSWindow  *window,
-                GSManager *manager)
+window_map_cb (GSWindow  *window,
+               GSManager *manager)
+{
+        gs_debug ("Handling window map event");
+}
+
+static void
+window_unmap_cb (GSWindow  *window,
+                 GSManager *manager)
+{
+        gs_debug ("window unmapped!");
+}
+
+static void
+manager_show_window (GSManager *manager,
+                     GSWindow  *window)
 {
         GSJob      *job;
         const char *theme;
 
-        g_return_if_fail (manager != NULL);
-        g_return_if_fail (GS_IS_MANAGER (manager));
-        g_return_if_fail (window != NULL);
-        g_return_if_fail (GS_IS_WINDOW (window));
-
         job = gs_job_new_for_widget (GTK_WIDGET (window));
 
         theme = select_theme (manager);
         gs_job_set_theme (job, theme, NULL);
 
-        if (! manager->priv->throttle_enabled) {
-                gs_job_start (job);
-        }
-
-        manager->priv->jobs = g_slist_append (manager->priv->jobs, job);
+        manager_add_job_for_window (manager, window, job);
 
         manager->priv->activate_time = time (NULL);
 
@@ -807,6 +930,20 @@ window_show_cb (GSWindow  *window,
         g_signal_emit (manager, signals [ACTIVATED], 0);
 }
 
+static void
+window_show_cb (GSWindow  *window,
+                GSManager *manager)
+{
+
+        g_return_if_fail (manager != NULL);
+        g_return_if_fail (GS_IS_MANAGER (manager));
+        g_return_if_fail (window != NULL);
+        g_return_if_fail (GS_IS_WINDOW (window));
+
+        gs_debug ("Handling window show");
+        manager_show_window (manager, window);
+}
+
 static void
 disconnect_window_signals (GSManager *manager,
                            GSWindow  *window)
@@ -815,7 +952,10 @@ disconnect_window_signals (GSManager *manager,
         g_signal_handlers_disconnect_by_func (window, window_dialog_up_cb, manager);
         g_signal_handlers_disconnect_by_func (window, window_dialog_down_cb, manager);
         g_signal_handlers_disconnect_by_func (window, window_show_cb, manager);
+        g_signal_handlers_disconnect_by_func (window, window_map_cb, manager);
         g_signal_handlers_disconnect_by_func (window, window_map_event_cb, manager);
+        g_signal_handlers_disconnect_by_func (window, window_unmap_cb, manager);
+        g_signal_handlers_disconnect_by_func (window, window_grab_broken_cb, manager);
 }
 
 static void
@@ -838,9 +978,15 @@ connect_window_signals (GSManager *manager,
         g_signal_connect_object (window, "dialog-down",
                                  G_CALLBACK (window_dialog_down_cb), manager, 0);
         g_signal_connect_object (window, "show",
-                                 G_CALLBACK (window_show_cb), manager, 0);
+                                 G_CALLBACK (window_show_cb), manager, G_CONNECT_AFTER);
+        g_signal_connect_object (window, "map",
+                                 G_CALLBACK (window_map_cb), manager, G_CONNECT_AFTER);
         g_signal_connect_object (window, "map_event",
-                                 G_CALLBACK (window_map_event_cb), manager, 0);
+                                 G_CALLBACK (window_map_event_cb), manager, G_CONNECT_AFTER);
+        g_signal_connect_object (window, "unmap",
+                                 G_CALLBACK (window_unmap_cb), manager, G_CONNECT_AFTER);
+        g_signal_connect_object (window, "grab_broken_event",
+                                 G_CALLBACK (window_grab_broken_cb), manager, G_CONNECT_AFTER);
 }
 
 static void
@@ -896,8 +1042,9 @@ gs_manager_create (GSManager *manager)
         }
         g_slist_free (manager->priv->windows);
 
-        for (i = 0; i < n_screens; i++)
+        for (i = 0; i < n_screens; i++) {
                 gs_manager_create_window (manager, gdk_display_get_screen (display, i));
+        }
 }
 
 GSManager *
@@ -920,6 +1067,17 @@ show_windows (GSList *windows)
         }
 }
 
+static void
+remove_job (GSJob *job)
+{
+        if (job == NULL) {
+                return;
+        }
+
+        gs_job_stop (job);
+        g_object_unref (job);
+}
+
 static gboolean
 gs_manager_activate (GSManager *manager)
 {
@@ -944,8 +1102,14 @@ gs_manager_activate (GSManager *manager)
                 return FALSE;
         }
 
-        if (! manager->priv->windows)
+        if (! manager->priv->windows) {
                 gs_manager_create (GS_MANAGER (manager));
+        }
+
+        manager->priv->jobs = g_hash_table_new_full (g_direct_hash,
+                                                     g_direct_equal,
+                                                     NULL,
+                                                     (GDestroyNotify)remove_job);
 
         manager->priv->active = TRUE;
 
@@ -985,12 +1149,7 @@ gs_manager_deactivate (GSManager *manager)
 
         gs_grab_release_keyboard_and_mouse ();
 
-        for (l = manager->priv->jobs; l; l = l->next) {
-                gs_job_stop (GS_JOB (l->data));
-                g_object_unref (l->data);
-        }
-        g_slist_free (manager->priv->jobs);
-        manager->priv->jobs = NULL;
+        manager_stop_jobs (manager);
 
         for (l = manager->priv->windows; l; l = l->next) {
                 gs_window_destroy (l->data);
@@ -1012,10 +1171,11 @@ gs_manager_set_active (GSManager *manager,
 {
         gboolean res;
 
-        if (active)
+        if (active) {
                 res = gs_manager_activate (manager);
-        else
+        } else {
                 res = gs_manager_deactivate (manager);
+        }
 
         return res;
 }
@@ -1040,8 +1200,9 @@ gs_manager_request_unlock (GSManager *manager)
         g_return_if_fail (manager != NULL);
         g_return_if_fail (GS_IS_MANAGER (manager));
 
-        if (! manager->priv->active)
+        if (! manager->priv->active) {
                 return;
+        }
 
         /* Find the screen that contains the pointer */
         display = gdk_display_get_default ();
